from .Button import Button
from .Slider import Slider
from .ColorPicker import ColorPicker
from .DatePicker import DatePicker
from .Progress import Progress
from .Radio import Radio
from .IconClipboard import IconClipboard
from .Switch import Switch
from .SliderFloat import SliderFloat
from .Label import Label
from .DayCalendar import DayCalendar
from .Toggle import Toggle
from .MultiSwitch import MultiSwitch
from .Tabs import Tabs
from .TextList import TextList
from .RangeSlider import RangeSlider
# from .Menu import Menu

from .deprecated import button
from .deprecated import slider
from .deprecated import colorPicker
from .deprecated import datePicker
from .deprecated import progress
from .deprecated import radio
from .deprecated import switch
from .deprecated import sliderFloat
from .deprecated import label
from .deprecated import dayCalendar
from .deprecated import toggle
from .deprecated import multiSwitch
from .deprecated import tabs
from .deprecated import textlist
from .deprecated import rangeSlider
# from .deprecated import menu


"""Map widget for interactive display of geospatial datasets"""
# Author(s): Davide.De-Marchi@ec.europa.eu, Edoardo.Ramalli@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, HTML, Layout, CallbackDispatcher
import ipyleaflet
from ipyleaflet import SearchControl, ScaleControl, FullScreenControl, WidgetControl, DrawControl, Icon
import ipyvuetify as v
from threading import Timer

# Vois imports
from vois.vuetify import settings, Toggle, Switch
from vois.geo import mapUtils
from vois.templates import PageConfigurator

# Name of layers
LAYERNAME_BACKGROUND = 'Background'
LAYERNAME_LABELS = 'Labels'


#####################################################################################################################################################
# Map class
#####################################################################################################################################################
class Map(ipyleaflet.Map):
    """
    Map widget for interactive display of geospatial datasets. This class inherits from the `ipyleaflet.Map class <https://ipyleaflet.readthedocs.io/en/latest/map_and_basemaps/map.html>`_ and adds methods and controls to simplify the Map customisation.

    Parameters
    ----------
    width : str, optional
        Width of the Map widget (default is '100%').
    height : str, optional
        Height of the Map widget (default is '600px').
    show_fullscreen : bool, optional
        Show or Hide the fullscreen control (default is True).
    show_search : bool, optional
        Show or Hide the Search control (default is True).
    show_scale : bool, optional
        Show or Hide the Scale control (default is True).
    show_coordinates : bool, optional
        Show or hide the Coordinates control (default is True).
    show_overview : bool, optional
        Show or hide the Overview control (default is False).
    show_basemaps : bool, optional
        Show or hide the Basemaps toggle control (default is True).
    color_first : str, optional
        Main color of the overlapped widgets (default is settings.color_first).
    color_second : str, optional
        Secondary color of the overlapped widgets (default is settings.color_second).
    dark : bool, optional
        Dark flag (default is settings.dark_mode).
    basemapindex : int, optional
        Initial basemap index (0=EC, 1=Esri, 2=Google), default is 0.
    onclick : python function, optional
        Callback function to call on click (receives as parameter: map, lon, lat, zoom), default is None.

    Example
    -------
    Creation of a map instance displaying an overview window::
        
        from IPython.display import display
        from vois.geo import Map

        m = Map(show_overview=True, center=[52,12], zoom=4, basemapindex=1)
        display(m)

    .. figure:: figures/Map.png
       :scale: 70 %
       :alt: Map widget

       Example of a Map with overview
    """

    # Initialization
    def __init__(self,
                 width='100%',
                 height='600px',
                 center=[50, 12],
                 zoom=5,
                 show_fullscreen=True,  # Show or Hide the fullscreen control
                 show_search=True,  # Show or Hide the Search control
                 show_scale=True,  # Show or Hide the Scale control
                 show_coordinates=True,  # Show or hide the Coordinates control
                 show_overview=False,  # Show or hide the Overview control
                 show_basemaps=True,  # Show or hide the Basemaps toggle control
                 color_first=None,  # Main color
                 color_second=None,  # Secondary color
                 dark=None,  # Dark flag
                 basemapindex=0,  # Initial basemap index (0=EC, 1=Esri, 2=Google)
                 onclick=None,  # Callback function to call on click (receives as parameter: map, lon, lat, zoom)
                 **kwargs):

        self._width = width
        self._height = height
        self._show_fullscreen = show_fullscreen
        self._show_search = show_search
        self._show_scale = show_scale
        self._show_coordinates = show_coordinates
        self._show_overview = show_overview
        self._show_basemaps = show_basemaps
        self._onclick = onclick

        self._color_first = color_first
        if self._color_first is None:
            self._color_first = settings.color_first

        self._color_second = color_second
        if self._color_second is None:
            self._color_second = settings.color_second

        self._dark = dark
        if self._dark is None:
            self._dark = settings.dark_mode

        self._basemapindex = basemapindex

        # Initial center and zoom of the map
        self.center = center
        self.zoom = zoom

        # List of WKT strings edited using the DrawControl
        self.wktstrings = []

        # DrawControl instance
        self._drawctrl = None

        # Card containing the widgets to configure the map appearance
        self.configure_card = None
        self.s1 = self.s2 = self.s3 = self.s4 = self.s5 = self.s6 = None
        self.update_properties = True

        # Map widget
        super().__init__(max_zoom=21, center=self.center, zoom=self.zoom, scroll_wheel_zoom=True,
                         basemap=mapUtils.EmptyBasemap(), attribution_control=False,
                         layout=Layout(width=self._width, height=self._height, margin='0px 0px 0px 0px'),
                         **kwargs)

        mapUtils.addLayer(self, mapUtils.OSM_EC(), LAYERNAME_BACKGROUND)
        mapUtils.addLayer(self, mapUtils.CartoLabels(), LAYERNAME_LABELS)
        layer = mapUtils.getLayer(self, LAYERNAME_LABELS)
        layer.opacity = 0.0
        self.onSelectBasemap(self._basemapindex)

        self.toggleBasemap = None
        self.cardCoordinates = None

        # Add FullScreen control
        self.show_fullscreen = self._show_fullscreen

        # Add Search control
        self.show_search = self._show_search

        # Add Scale control
        self.show_scale = self._show_scale

        # Add widget control to select basemaps
        self.show_basemaps = self._show_basemaps

        # Add overview
        self.show_overview = self._show_overview

        # Add widget control to display geographic coordinates at mouse move
        self.show_coordinates = self._show_coordinates

        # Manage interaction events
        self._interaction_callbacks = CallbackDispatcher()
        self.on_interaction(self.handleMapInteraction)

    # Fake "draw" method
    def draw(self):
        return self

    # Remove all layers from map
    def clear(self):
        """
        Remove all the additional layers from the map.
        """
        mapUtils.clear(self)
        mapUtils.addLayer(self, mapUtils.EmptyBasemap(), LAYERNAME_BACKGROUND)
        mapUtils.addLayer(self, mapUtils.CartoLabels(), LAYERNAME_LABELS)
        self.onSelectBasemap(self.basemapindex)

    # Add a ipyleaflet.TileLayer to the map
    def addLayer(self, tileLayer, name=None, opacity=1.0):
        """
        Add a new layer to the map.
        
        Parameters
        ----------
        tileLayer : ipyleaflet.TileLayer or a class that has a tileLayer() callable
            Layer to add to the map
        name : str, optional
            Name of the layer (default is None)
        opacity : float, optional
            Opacity of the layer in the range [0.0,1.0] (default is 1.0)

        Returns
        -------
        layer : instance of ipyleafler.Layer
            The layer added to the map
        """

        # if no name is passed, generate a layer name
        if name is None:
            count = len(self.layers)
            name = "layer%d" % (count + 1)

        # In case a vectorlayer or rasterlayer is passed: call .tileLayer() to get the ipyleaflet.TileLayer instance!
        if isinstance(tileLayer, ipyleaflet.TileLayer):
            res = mapUtils.addLayer(self, tileLayer, name=name, opacity=opacity)
        else:
            res = mapUtils.addLayer(self, tileLayer.tileLayer(), name=name, opacity=opacity)

        return res

    # Manage all user interaction on the map
    def handleMapInteraction(self, **kwargs):

        if self._show_coordinates:
            if kwargs.get('type') == 'mousemove':
                lon = kwargs.get('coordinates')[1]
                lat = kwargs.get('coordinates')[0]
                self.cardCoordinates.children = [
                    v.Html(tag='div', children=[' %.4f° N, %.4f° E' % (lat, lon)], style_='color: black;',
                           class_='pa-0 ma-0 ml-1 mr-1')]

        if self._onclick is not None:
            if kwargs.get('type') == 'click':
                lon = kwargs.get('coordinates')[1]
                lat = kwargs.get('coordinates')[0]
                self._onclick(self, lon, lat, int(self.zoom))

    # Selection of a basemap
    def onSelectBasemap(self, index):
        layer = mapUtils.getLayer(self, LAYERNAME_LABELS)

        self._basemapindex = max(0, min(2, index))

        if self._basemapindex == 1:
            basemaplayer = mapUtils.EsriWorldImagery()
            layer.opacity = 1.0
        elif self._basemapindex == 2:
            basemaplayer = mapUtils.GoogleHybrid()
            layer.opacity = 0.0
        else:
            basemaplayer = mapUtils.OSM_EC()
            layer.opacity = 0.0

        mapUtils.addLayer(self, basemaplayer, LAYERNAME_BACKGROUND)

    # Change of configure widgets
    def change_fullscreen(self, flag):
        if self.update_properties:
            self.show_fullscreen = flag

    def change_coordinates(self, flag):
        if self.update_properties:
            self.show_coordinates = flag

    def change_search(self, flag):
        if self.update_properties:
            self.show_search = flag

    def change_scale(self, flag):
        if self.update_properties:
            self.show_scale = flag

    def change_basemaps(self, flag):
        if self.update_properties:
            self.show_basemaps = flag

    def change_overview(self, flag):
        if self.update_properties:
            self.show_overview = flag

    # Create a card instance to configure the Map appearance and returns it
    def configure(self):
        if self.configure_card is None:
            self.configure_card = v.Card(flat=True, class_='pa-2 ma-0')

            self.s1 = Switch(self.show_fullscreen, 'Add the Fullscreen control', inset=True, dense=True,
                             on_change=self.change_fullscreen, color=self._color_first)
            self.s2 = Switch(self.show_coordinates, 'Add the Coordinates control', inset=True, dense=True,
                             on_change=self.change_coordinates, color=self._color_first)
            self.s3 = Switch(self.show_search, 'Add the Search control', inset=True, dense=True,
                             on_change=self.change_search, color=self._color_first)
            self.s4 = Switch(self.show_scale, 'Add the Scale control', inset=True, dense=True,
                             on_change=self.change_scale, color=self._color_first)
            self.s5 = Switch(self.show_basemaps, 'Add the Basemaps control', inset=True, dense=True,
                             on_change=self.change_basemaps, color=self._color_first)
            self.s6 = Switch(self.show_overview, 'Add the Overview control', inset=True, dense=True,
                             on_change=self.change_overview, color=self._color_first)

            self.configure_card.children = [widgets.VBox(
                [PageConfigurator.label('Map', color='black'), self.s1, self.s2, self.s3,
                 self.s4, self.s5, self.s6])]

        return self.configure_card

    @property
    def content(self):
        return 'Map'

    @content.setter
    def content(self, c):
        pass

    @property
    def width(self):
        """
        Get/Set the width of the Map widget.
        """
        return self._width

    @width.setter
    def width(self, w):
        self._width = w
        self.layout.width = self._width

    @property
    def height(self):
        """
        Get/Set the height of the Map widget.
        """
        return self._height

    @height.setter
    def height(self, h):
        self._height = h
        self.layout.height = self._height

    @property
    def show_fullscreen(self):
        """
        Display or hides the Fullscreen control.
        """
        return self._show_fullscreen

    @show_fullscreen.setter
    def show_fullscreen(self, flag):
        self._show_fullscreen = flag

        if self.s1 is not None:
            self.update_properties = False
            self.s1.value = flag
            self.update_properties = True

        for control in self.controls:
            if isinstance(control, ipyleaflet.leaflet.FullScreenControl):
                self.remove(control)

        if self._show_fullscreen:
            self.add(FullScreenControl(position="topright"))

    @property
    def show_coordinates(self):
        """
        Display or hides the Coordinates control.
        """
        return self._show_coordinates

    @show_coordinates.setter
    def show_coordinates(self, flag):
        self._show_coordinates = flag

        if self.s2 is not None:
            self.update_properties = False
            self.s2.value = flag
            self.update_properties = True

        mapUtils.removeCoordinates(self)

        if self._show_coordinates:
            self.cardCoordinates = mapUtils.getCoordinatesCard(self)

    @property
    def show_search(self):
        """
        Display or hides the Search control.
        """
        return self._show_search

    @show_search.setter
    def show_search(self, flag):
        self._show_search = flag

        if self.s3 is not None:
            self.update_properties = False
            self.s3.value = flag
            self.update_properties = True

        for control in self.controls:
            if isinstance(control, ipyleaflet.leaflet.SearchControl):
                self.remove(control)

        if self._show_search:
            self.add(
                SearchControl(position="topleft", url='https://nominatim.openstreetmap.org/search?format=json&q={s}',
                              zoom=12))

    @property
    def show_scale(self):
        """
        Display or hides the Scale control.
        """
        return self._show_scale

    @show_scale.setter
    def show_scale(self, flag):
        self._show_scale = flag

        if self.s4 is not None:
            self.update_properties = False
            self.s4.value = flag
            self.update_properties = True

        for control in self.controls:
            if isinstance(control, ipyleaflet.leaflet.ScaleControl):
                self.remove(control)

        if self._show_scale:
            self.add(ScaleControl(position='topleft'))

    @property
    def show_basemaps(self):
        """
        Display or hides the Basemaps selection control.
        """
        return self._show_basemaps

    @show_basemaps.setter
    def show_basemaps(self, flag):
        self._show_basemaps = flag

        if self.s5 is not None:
            self.update_properties = False
            self.s5.value = flag
            self.update_properties = True

        mapUtils.removeCardByName(self, 'Basemaps')

        if self._show_basemaps:
            c = mapUtils.getCardByName(self, 'Basemaps', position='bottomright')
            c.tile = True
            self.toggleBasemap = Toggle(self._basemapindex,
                                        ['Gisco', 'Esri', 'Google'],
                                        tooltips=['Select EC Gisco roadmap as background layer',
                                                  'Select ESRI WorldImagery as background layer',
                                                  'Select GOOGLE Satellite as background layer'],
                                        color_selected=self._color_first, color_unselected=self._color_second,
                                        rounded=False,
                                        dark=self._dark, on_change=self.onSelectBasemap, row=True, width=70,
                                        justify='start', padding_row=0, tile=True)
            c.children = [self.toggleBasemap.draw()]

    @property
    def show_overview(self):
        """
        Display or hides the Overview control.
        """
        return self._show_overview

    @show_overview.setter
    def show_overview(self, flag):
        self._show_overview = flag
        mapUtils.removeOverview(self)

        if self.s6 is not None:
            self.update_properties = False
            self.s6.value = flag
            self.update_properties = True

        if self._show_overview:
            mapUtils.addOverview(self, position='bottomleft')

    @property
    def color_first(self):
        """
        Get/Set the primary color.
        """
        return self._color_first

    @color_first.setter
    def color_first(self, color):
        self._color_first = color

        if self.toggleBasemap is not None:
            self.toggleBasemap.colorselected = self._color_first

        if self.configure_card is not None:
            self.s1.color = self._color_first
            self.s2.color = self._color_first
            self.s3.color = self._color_first
            self.s4.color = self._color_first
            self.s5.color = self._color_first
            self.s6.color = self._color_first

    @property
    def color_second(self):
        """
        Get/Set the secondary color.
        """
        return self._color_second

    @color_second.setter
    def color_second(self, color):
        self._color_second = color

        if self.toggleBasemap is not None:
            self.toggleBasemap.colorunselected = self._color_second

    @property
    def dark(self):
        """
        Get/Set the dark mode flag.
        """
        return self._dark

    @dark.setter
    def dark(self, flag):
        self._dark = flag

        if self.toggleBasemap is not None:
            self.toggleBasemap.dark = self._dark

    @property
    def basemapindex(self):
        """
        Get/Set the current basemap (index from 0 to 1).
        """
        return self._basemapindex

    @basemapindex.setter
    def basemapindex(self, index):
        self.onSelectBasemap(index)

        if self.toggleBasemap is not None:
            self.toggleBasemap.value = self._basemapindex

    @property
    def state(self):
        return {x: getattr(self, x) for x in ['content',
                                              # 'width',      # Will inherit from content!!!
                                              # 'height',
                                              'show_fullscreen',
                                              'show_coordinates',
                                              'show_search',
                                              'show_scale',
                                              'show_basemaps',
                                              'show_overview',
                                              'color_first',
                                              'color_second',
                                              'dark',
                                              'basemapindex',
                                              'center',
                                              'zoom'
                                              ]}

    @state.setter
    def state(self, statusdict):
        for key, value in statusdict.items():
            setattr(self, key, value)

    @property
    def onclick(self):
        """
        Get/Set the python function to call when the user clicks on the map.
        """
        return self._onclick

    @onclick.setter
    def onclick(self, callback):
        self._onclick = callback

    #####################################################################################################################################################
    # Feature draw
    #####################################################################################################################################################

    @property
    def drawctrl(self):
        """
        Display or hides the Feature Draw control. The wktstrings member of the Map class contains the list of all the features added in WKT format.
        """
        if self._drawctrl is None:
            return False
        else:
            return True

    @drawctrl.setter
    def drawctrl(self, flag):
        if flag:
            if self._drawctrl is None:
                self._drawctrl = DrawControl()

                self._drawctrl.marker = {
                    "shapeOptions": {
                        "color": self._color_first,
                        "weight": 4,
                        "opacity": 1.0
                    }
                }

                self._drawctrl.polyline = {
                    "shapeOptions": {
                        "color": self._color_first,
                        "weight": 4,
                        "opacity": 1.0
                    }
                }

                self._drawctrl.polygon = {
                    "shapeOptions": {
                        "fillColor": self._color_first,
                        "color": self._color_first,
                        "fillOpacity": 0.2
                    },
                    "drawError": {
                        "color": "#dd253b",
                        "message": "Error!"
                    },
                    "allowIntersection": True
                }

                self._drawctrl.circle = {}

                self._drawctrl.circlemarker = {}

                self._drawctrl.rectangle = {
                    "shapeOptions": {
                        "fillColor": self._color_first,
                        "color": self._color_first,
                        "fillOpacity": 0.2
                    }
                }

                self._drawctrl.on_draw(callback=self.on_draw)

                self.add(self._drawctrl)
        else:
            if self._drawctrl is not None:
                self.remove(self._drawctrl)
                self._drawctrl = None

    def updateWKT(self):
        self.wktstrings = [geojson2WKT(x) for x in self._drawctrl.data]

    def on_draw(self, control, action, geo_json):
        t = Timer(1.0, self.updateWKT)
        t.start()


#####################################################################################################################################################
# Utility: convert geojson to WKT
#####################################################################################################################################################
def geojson2WKT(geojson):
    # Force a longitude in [0,360]
    def lon(x):
        while x > 360: x -= 360
        while x < 0:   x += 360
        return x

    # Force a latitude in [0,90]
    def lat(y):
        while y > 90: y -= 90
        while y < 0:   y += 90
        return y

    # From a list of x,y to lon,lat
    def c2latlon(c):
        return lon(c[0]), lat(c[1])

    coords = geojson['geometry']

    wkt = coords['type'].upper() + '('

    if coords['type'] == 'Point':
        c = coords['coordinates']
        wkt += '%f %f' % c2latlon(c)
    elif coords['type'] == 'LineString':
        wkt += ','.join(['%f %f' % c2latlon(x) for x in coords['coordinates']])
    else:
        for ring in coords['coordinates']:
            wkt += '(' + ','.join(['%f %f' % c2latlon(x) for x in ring]) + ')'
            break

    wkt += ')'
    return wkt



"""Slider to select a range of numeric values."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import warnings
import ipyvuetify as v

from vois.vuetify.utils.util import *
from typing import Callable, Any, Union, Optional


#####################################################################################################################################################
# Integer range slider class
#####################################################################################################################################################
class RangeSlider(v.RangeSlider):
    """
    Slider to select a range of numeric values.
        
    Parameters
    ----------
    selected_min_value : numeric
        Minimum value of the selected range
    selected_max_value : numeric
        Maximum value of the selected range
    min_value : numeric
        Minimum value selectable by the user
    max_value : numeric
        Maximum value selectable by the user
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    onchange : function, optional
        Python function to call when the user selects a value. The function will receive two parameters of numeric type containing the current min and max value selected by the user
    height : int, optional
        Height of the slider widget in pixel (default is 250 pixels)
    vertical : bool, optional
        Flag to display the range slider in vertical mode (default is True)
            
    Example
    -------
    Creation and display of a range slider widget::
        
        from vois.vuetify import RangeSlider
        
        s = RangeSlider(5,18, 0,20)
        display(s)

    .. figure:: figures/rangeSlider.png
       :scale: 100 %
       :alt: rangeSlider widget

       RangeSlider widget example
   """

    deprecation_alias = dict(selectedminvalue='selected_min_value', selectedmaxvalue='selected_max_value',
                             minvalue='min_value', maxvalue='max_value', onchange='on_change')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 selected_min_value: int,
                 selected_max_value: int,
                 min_value: int,
                 max_value: int,
                 color: Optional[str] = None,
                 on_change: Optional[Callable[[tuple[int, int]], None]] = None,
                 height: int = 250,
                 vertical: bool = True):

        from vois.vuetify import settings

        self.selected_min_value = selected_min_value
        self.selected_max_value = selected_max_value
        self.min_value = min_value
        self.max_value = max_value
        self.color = color if color is not None else settings.color_first
        self.on_change = on_change
        self.height = height
        self.vertical = vertical

        if self.vertical:
            margins = "ml-n5 mr-1 mt-n6 mb-n7"
        else:
            margins = "ml-5  mr-5 mt-4  mb-n6"

        self.slider = self

        super().__init__(v_model=[self.selected_min_value, self.selected_max_value], dense=True, small=True,
                         thumb_color=self.color,
                         thumb_label="always", thumb_size=32, ticks=True, ticks_size=10,
                         color=self.color, track_color="grey", class_="pa-0 ma-0 %s" % margins,
                         min=self.min_value, max=self.max_value, vertical=self.vertical, height=self.height)

        # If requested onchange management
        if not self.on_change is None:
            self.slider.on_event('end', self.__internal_onchange)

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Manage onchange event
    def __internal_onchange(self, widget=None, event=None, data=None):
        if self.on_change:
            self.on_change(data[0], data[1])

    # Returns the vuetify object to display (the v.Container)
    def draw(self):
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self



"""Label widget to display a text with an optional icon."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v

from vois.vuetify import tooltip, fontsettings
from vois.vuetify.utils.util import *

from typing import Callable, Optional, Any


#####################################################################################################################################################
# Label class that can display also an icon
#####################################################################################################################################################
class Label(v.Html):
    """
    Label widget to display a text with an optional icon.
        
    Parameters
    ----------
    text : str
        Test string to be displayed on the label widget
    onclick : function, optional
        Python function to call when the user clicks on the label. The function will receive as parameter the value of the argument (default is None)
    argument : any, optional
        Argument to be passed to the onclick function when user click on the label (default is None)
    disabled : bool, optional
        Flag to show the label as disabled (default is False)
    text_weight : int, optional
        Weight of the text to be shown in the label (default is 350, Bold is any value greater or equal to 500)
    height : int, optional
        Height of the label widget in pixels (default is 20)
    margins : int, optional
        Dimension of the margins on all directions (default is 0)
    margin_top : int, optional
        Dimension of the margin on top of the label (default is None)
    icon: str, optional
        Name of the icon to display aside the text of the label (default is None)
    icon_large : bool, optional
        Flag that sets the large version of the icon (default is False)
    icon_small : bool, optional
        Flag that sets the small version of the icon (default is False)
    icon_left : bool, optional
        Flag that sets the position of the icon  to the left of the text of the label (default is False)
    icon_color : str, optional
        Color of the icon (default is 'black')
    tooltip_text : str, optional
        Tooltip string to display when the user hovers on the label (default is None)
    text_color : str, optional
        Color used for the label text
    back_color : str, optional
        Color used for the background of the label
    dark : bool, optional
        Flag to invert the text and back_color (default is the value of settings.dark_mode)
            

    Note
    ----
    All the icons from https://materialdesignicons.com/ site can be used, just by prepending 'mdi-' to their name.
    
    All the free icons from https://fontawesome.com/ site can be used, just by prepending 'fa-' to their name.


    Example
    -------
    Creation and display of a label widget containing an icon::
        
        from vois.vuetify import Label

        lab = Label('Test label', text_weight=300, margins=2,
                      icon='mdi-car-light-high', icon_color='red',
                      icon_large=True, height=22)

        display(lab)
    
    
    .. figure:: figures/label.png
       :scale: 100 %
       :alt: label widget

       Example of a label widget with text and an icon.
   """

    # Initialization
    deprecation_alias = dict(textweight='text_weight', margintop='margin_top', iconlarge='icon_large',
                             iconsmall='icon_small', iconleft='icon_left', iconcolor='icon_color',
                             textcolor='text_color', onclick='on_click', backcolor='back_color')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 text: str,
                 on_click: Optional[Callable[[Optional[Any]], None]] = None,
                 argument: Optional[Any] = None,
                 disabled: bool = False,
                 text_weight: int = 350,
                 height: int = 20,
                 margins: int = 0,
                 margin_top: Optional[int] = None,
                 icon: Optional[str] = None,
                 icon_large: bool = False,
                 icon_small: bool = False,
                 icon_left: bool = False,
                 icon_color: str = 'black',
                 tooltip_text: Optional[str] = None,
                 text_color: Optional[str] = None,
                 back_color: Optional[str] = None,
                 dark: Optional[str] = None,
                 **kwargs):

        super().__init__(tag='div', **kwargs)

        from vois.vuetify import settings

        self.labeltext = text
        self.on_click = on_click
        self.argument = argument

        self.disabled = disabled
        self.text_weight = text_weight
        self.height = height
        self.margins = margins
        self.margin_top = margin_top

        self.icon = icon
        self.icon_left = icon_left
        self.icon_large = icon_large
        self.icon_small = icon_small
        self.icon_color = icon_color

        self.tooltip_text = tooltip_text

        self.text_color = text_color
        self.back_color = back_color
        self.dark = dark if dark is not None else settings.dark_mode

        if not self.dark is None:
            if self.dark:
                if self.back_color is None:
                    self.back_color = '#111111'
                if self.text_color is None:
                    self.text_color = 'white'

        self.__createLabel()

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Create the label
    def __createLabel(self):
        if self.icon is None:
            childs = [self.labeltext]
        else:
            childs = [self.labeltext,
                      v.Icon(left=self.icon_left, large=self.icon_large, small=self.icon_small, color=self.icon_color,
                             children=[self.icon])]

        strstyle = 'font-family: %s; font-size: 10; font-weight: %d; text-transform: none;' % (
            fontsettings.font_name, self.text_weight)
        if not self.text_color is None:
            strstyle += 'color: %s;' % self.text_color
        if not self.back_color is None:
            strstyle += 'background-color: %s;' % self.back_color

        self.item = v.Card(disabled=self.disabled, elevation=0, height=self.height, depressed=True, children=childs,
                           style_=strstyle)
        if not self.tooltip_text is None:
            self.item = tooltip.tooltip(self.tooltip_text, self.item)

        # If requested onclick management
        if not self.on_click is None:
            self.item.on_event('click', self.__internal_onclick)

        if not self.margin_top is None:
            self.container = v.Container(class_="pa-0 ma-%s mt-%s" % (str(self.margins), str(self.margin_top)),
                                         children=[self.item])
        else:
            self.container = v.Container(class_="pa-0 ma-%s" % (str(self.margins)), children=[self.item])

        self.children = [self.container]

    # Manage click event
    def __internal_onclick(self, widget=None, event=None, data=None):
        if self.on_click:
            if not self.argument is None:
                self.on_click(self.argument)
            else:
                self.on_click()

    # Returns the vuetify object to display (the v.Container)
    def draw(self):
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self

    # Get the label text
    @property
    def text(self):
        """
        Get/Set the label text.
        
        Returns
        --------
        text : str
            Text currently shown in the label

        Example
        -------
        Programmatically set the label text (needs a re-display to be visible!)::
            
            lab.text = 'New text for the label'
            display(lab.draw())
            print(lab.text)
        
        """
        return self.labeltext

    # Set the label text
    @text.setter
    def text(self, new_text):
        self.labeltext = new_text
        self.__createLabel()



"Geojson layer class to be displayed in a ipyleaflet.Map with selection management"
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2025
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import CallbackDispatcher
import ipyleaflet


#####################################################################################################################################################
# geojsonLayer class (fully tested for point features but almost ready for extention to polygons and polylines)
#####################################################################################################################################################
class geojsonLayer():
    
    # Initialization
    def __init__(self,
                 featureCollection,
                 style={},
                 point_style={'radius': 9, 'fillColor': "#ff0000", 'color': "#000000", 'weight': 1.5, 'opacity': 1.0, 'fillOpacity': 0.8},
                 style_callback=None,
                 hover_style={'fillColor': 'orange' , 'fillOpacity': 0.7, 'dashArray': '3', 'weight': 2.0 },
                 selection_style={},
                 selection_point_style={'radius': 10, 'color': "#ffff00", 'weight': 4.0, 'opacity': 1.0, 'fillOpacity': 0.0},
                 name='Layer',
                 on_click=None):
        
        # Feature collection to display
        self.featureCollection = featureCollection
                 
        # Styling members
        self.style                 = style
        self.point_style           = point_style
        self.style_callback        = style_callback
        self.hover_style           = hover_style
        self.selection_style       = selection_style
        self.selection_point_style = selection_point_style
        
        # Callback 
        self.on_click = on_click
       
        
        # GeoJSON layer
        self.geojson = ipyleaflet.GeoJSON(data=self.featureCollection, point_style=self.point_style, style=self.style, hover_style=self.hover_style, name=name)
        if not style_callback is None:
            self.geojson.style_callback = self.style_callback
        
        self.geojson.on_click(self.__internal_on_click)
        
        
        # Feature collection to manage the selection
        self.selection = { "type": "FeatureCollection", "crs": { "type": "name", "properties": { "name": "urn:ogc:def:crs:OGC:1.3:CRS84" } },
                           "features": [ { "type": "Feature", "properties": { "name": "selected" }, "geometry": { "type": "Point", "coordinates": [ 99999, 99999 ] } } ]
                        }
        
        self.selection_geojson = ipyleaflet.GeoJSON(data=self.selection, point_style=self.selection_point_style, style=self.selection_style, hover_style=self.hover_style)

        
        # Map where the layer is added
        self.m = None

        # True if a click on the map must be skipped
        self.skip_map_click = False        
        
        
    # Add the layer to an ipyleaflet Map and manage the click to unselect a geojson feature
    def addToMap(self, m):
        self.m = m
        self.m.add(self.geojson)
        self.m.add(self.selection_geojson)
        
        self.skip_map_click = False
        self.m._interaction_callbacks = CallbackDispatcher()
        self.m.on_interaction(self.handle_interaction_map)
        

    # Handle click on the map (to unselect)
    def handle_interaction_map(self, **kwargs):
        if kwargs.get('type') == 'click':
            if not self.skip_map_click:
                self.selection['features'] = [ { "type": "Feature", "properties": { "name": "selected" }, "geometry": { "type": "Point", "coordinates": [ 99999, 99999 ] } }  ]
                newsel = ipyleaflet.GeoJSON(data=self.selection, point_style=self.selection_point_style, style=self.selection_style, hover_style=self.hover_style)
                self.m.substitute(self.selection_geojson,newsel)
                self.selection_geojson = newsel
                if not self.on_click is None:
                    self.on_click(None)
            self.skip_map_click = False
    
    
    # Select one of the features
    def selectFeature(self, feature):
        self.selection['features'] = [ { "type": "Feature", "properties": { "name": "selected" }, "geometry": feature['geometry'] }  ]
        newsel = ipyleaflet.GeoJSON(data=self.selection, point_style=self.selection_point_style, style=self.selection_style)
        self.m.substitute(self.selection_geojson,newsel)
        self.selection_geojson = newsel
        
        
    # Handle click on a geojson feature to select
    def __internal_on_click(self, event, feature, properties, id):
        if event == 'click':
            self.skip_map_click = True
            self.selectFeature(feature)
            if not self.on_click is None:
                self.on_click(feature)

        
    


"""Upload local json file: dialog box to select and preview of some fields"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, Layout
import ipyvuetify as v

# Vois imports
from vois import cssUtils, colors
from vois.vuetify import settings, upload, dialogGeneric, dialogMessage, dialogWait

# Python imports
import json


# Creation of a label
def label(text, class_='pa-0 ma-0 mt-1 mr-3', size=14, weight=400, color=settings.color_first, width=None):
    lab = v.Html(tag='div', children=[text], class_=class_)
    if width is None:
        lab.style_ = 'font-size: %dpx; font-weight: %d; color: %s;'%(size,weight,color)
    else:
        lab.style_ = 'font-size: %dpx; font-weight: %d; color: %s; width: %dpx'%(size,weight,color,int(width))
    return lab


#####################################################################################################################################################
# Upload local json file: dialog box to select and previw of some fields
#####################################################################################################################################################
class UploadJson():
    """
    Upload local json file: dialog box to select and preview of some fields. Widget that simplifies the selection of a json from the local computer to be uploaded in the cloud. It features the preview of one or more field of the json into a modal dialog-box.

    Parameters
    ----------
    output : ipywidgets.Output
        Output widget on which the widget has to be displayed
    message : str, optional
        Message to display in the upload widget to guide the user (default is 'Click to select the file to upload').
    label : str, optional
        Placeholder label displayed on top of the upload widget (default is 'File json:').
    accept : str, optional
        List of mime types accepted for the selection (default is 'application/json').
    title : str, optional
        Title of the modal dialog-box (default is 'Select a file').
    required_attributes : list of str, optional
        List of attribute names to check in selected json file and to display in the preview (default is [])
    attributes_width : int, optional
        Width in pixel of the space to display the required attributes names (default is 100)
    color : str, optional
        Main color of the widgets (default is settings.color_first).
    dark : bool, optional
        Dark flag (default is settings.dark_mode).
    titleheight : int, optional
        Height in pixels of the title bar of the modal dialog-box (default is 40)
    width : int, optional
        Width in pixels of the modal dialog-box (default is 800)
    height : int, optional
        height in pixels of the preview area of the modal dialog-box (default is 200)
    onOK : function, optional
        Python function to call when the user clicks on the OK button. The function will receive as parameter the json dictionary read from the selected file.
    onCancel : function, optional
        Python function to call when the user clicks on the CANCEL button. The function will receive no parameters.
    
    Example
    -------
    Display of a dialog-box to enable the user to select a json file from its local machine::
        
        from vois.vuetify import UploadJson
        from ipywidgets import widgets, Layout
        from IPython.display import display
        import ipyvuetify as v

        output = widgets.Output(layout=Layout(width='0px', height='0px'))
        display(output)

        debug = widgets.Output()

        def onSelectedJson(jsonitem):
            debug.clear_output()
            with debug:
                display(jsonitem)

        u = UploadJson.UploadJson(output,
                                  onOK=onSelectedJson,
                                  required_attributes=['appname', 'title'],
                                  attributes_width=80)
        u.show()

        display(debug)

    .. figure:: figures/upload_json_1.png
       :scale: 50 %
       :alt: Upload json dialog widget

       UploadImage dialog box before the selection of a json file
       
    .. figure:: figures/upload_json_2.png
       :scale: 50 %
       :alt: Upload dialog widget with a json file selected

       UploadJson dialog box showing the preview of some fields of the selected json file
    """
    
    # Initialization
    def __init__(self,
                 output,
                 message='Click to select the file to upload',
                 label='File json:',
                 accept='application/json',
                 title='Select a file',
                 required_attributes=[],         # List of attribute names to check in selected json file and to display in the preview
                 attributes_width=100,           # width in pixel of the spavce to display the required attributes names
                 color=None,
                 dark=None,
                 titleheight=40,
                 width=800,
                 height=200,
                 onOK=None,                      # Called passing the content of the selected file
                 onCancel=None):                 # Called with no argument
        
        self.output              = output
        self.title               = title
        self.required_attributes = required_attributes
        self.attributes_width    = attributes_width
        self.titleheight         = titleheight
        self.width               = width
        self.height              = height
        self.onOK                = onOK
        self.onCancel            = onCancel
        
        self.color = color
        if self.color is None:
            self.color = settings.color_first
            
        self.dark = dark
        if self.dark is None:
            self.dark = settings.dark_mode
            
        cssUtils.allSettings(self.output)
        
        self.json = None
        
        self.wait = None
        
        self.preview = widgets.Output(layout=Layout(height='%dpx'%self.height))
        self.u = upload.upload(accept=accept, label=label, onchanging=self.onFileSelected, onchange=self.onFileUpload, placeholder=message, multiple=False, color=self.color)

        spacerY = v.Html(tag='div', style_='width: 0px; height: 20px;')
        self.content = v.Card(flat=True, class_='pa-0, ma-0 mt-n4 ml-4 mr-4', children=[widgets.VBox([self.u.draw(), spacerY, self.preview])])
        
        
    # Open the dialog-box
    def show(self):
        """
        Opens the dialog-box.
        """
        self.preview.clear_output()
        self.u.clear()
        self.dlg = dialogGeneric.dialogGeneric(title=self.title, on_ok=self._internal_onOK, on_cancel=self._internal_onCancel, on_close=self.onCancel,
                                               text='   ', color=self.color, dark=self.dark, titleheight=self.titleheight, width=self.width,
                                               show=True, addclosebuttons=True, addokcancelbuttons=True,
                                               fullscreen=False, content=[self.content], output=self.output)
        self.dlg.okdisabled = True
        
        
    # Just after the file is selected
    def onFileSelected(self):
        self.wait = dialogWait.dialogWait(text='Uploading file...', output=self.output, color=self.color, dark=self.dark)

        
    # Selection of an image to upload
    def onFileUpload(self, files):
        # If at least one file has been selected      
        if len(files) > 0:
            
            self.preview.clear_output()
            for f in files:
                self.json = json.load(f['file_obj'])
                
                # Check the presence of the required attributes
                attributesOK = True
                for a in self.required_attributes:
                    if a not in self.json:
                        attributesOK = False
                        break
                        
                if attributesOK:
                    self.dlg.okdisabled = False
                    with self.preview:
                        for a in self.required_attributes:
                            display(widgets.HBox([label(a+':', color='black', width=self.attributes_width), label(self.json[a], color='black', weight=600)]))
                        
                else:
                    dialogMessage.dialogMessage(title='Error',
                                                text='The uploaded file has no %s attribute'%a,
                                                addclosebuttons=True, show=True, width=400, output=self.output)
                    
                    self.dlg.okdisabled = True
                    self.json = None
                    self.preview.clear_output()
                    
        # No files selected
        else:
            self.json = None
            self.preview.clear_output()

        if self.wait is not None:
            self.wait.close()
    
    
    # Selection done
    def _internal_onOK(self):
        if self.onOK is not None:
            self.onOK(self.json)
    
    # Exit without selection
    def _internal_onCancel(self):
        if self.onCancel is not None:
            self.onCancel()



"""Widget to select a float value"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
#
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by
# the European Commission subsequent versions of the EUPL (the "Licence");
#
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
#
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
from ipywidgets import widgets
import ipyvuetify as v

from vois.vuetify.utils.util import *

from vois.vuetify import settings, tooltip
from typing import Callable, Any, Union, Optional


#####################################################################################################################################################
# sliderFloat control
#####################################################################################################################################################
class SliderFloat(v.Row):
    """
    Compound widget to select a float value in a specific range.

    Parameters
    ----------
    value : float, optional
        Initial value of the slider (default is 1.0)
    min_value : float, optional
        Minimum value of the slider (default is 0.0)
    max_value : float, optional
        Maximum value of the slider (default is 1.0)
    text : str, optional
        Text to display on the left of the slider (default is 'Select')
    show_percentage : bool, optional
        If True, the widget will write the current value as a percentage inside the allowed range and will append the % sign to the right of the current value (default is True)
    decimals : int, optional
        Number of decimal digits to use in case show_percentage is False (default is 2)
    max_int : int, optional
        Maximum integer number for the underlining integer slider (defines the slider sensitivity). Default value is None, meaning it will be automatically calculated
    label_width : int, optional
        Width of the label part of the widgets in pixels (default is 0, meaning it is automatically calculated based on the provided label text)
    slider_width : int, optional
        Width in pixels of the slider component of the widget (default is 200)
    reset_button: bool, optional
        If True a reset button is displayed, allowing for resetting the widget to its initial value (default is False)
    show_tooltip: bool, optional
        If True the up and down buttons will have a tooltip (default is False)
    on_change : function, optional
        Python function to call when the changes the value of the slider. The function will receive a single parameter, containing the new value of the slider in the range from minvalue to maxvalue (default is None)
    color : str, optional
        Color of the widget (default is settings.color_first)
    editable : bool, optional
        If True the label can be edited by clicking on it (default is False)
    editable_width : int, optional
        If the slider is editable, set the width of the v.TextField widget to enter the value (default is 90)

    Example
    -------
    Creation and display of a slider widget to select an opacity value in the range [0.0, 1.0]::

        from vois.vuetify import sliderFloat
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def on_change(value):
            with output:
                print(value)

        o = SliderFloat(0.8, text='Fill opacity:', min_value=0.0, max_value=1.0, on_change=on_change)

        display(o)
        display(output)

    .. figure:: figures/sliderFloat.png
       :scale: 100 %
       :alt: opacity widget

       Example of a slider widget to select a floating point value.
   """

    deprecation_alias = dict(showpercentage='show_percentage', minvalue='min_value', maxvalue='max_value',
                             onchange='on_change', maxint='max_int', labelwidth='label_width',
                             sliderwidth='slider_width', resetbutton='reset_button', showtooltip='show_tooltip',
                             editableWidth='editable_width')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 value: Union[int, float] = 1.0,
                 min_value: Union[int, float] = 0.0,
                 max_value: Union[int, float] = 1.0,
                 text: str = 'Select',
                 show_percentage: bool = True,
                 decimals: int = 2,
                 max_int: Optional[int] = None,
                 label_width: int = 0,
                 slider_width: int = 200,
                 reset_button: bool = False,
                 show_tooltip: bool = False,
                 on_change: Optional[Callable[[float], None]] = None,
                 color: str = None,
                 editable: bool = False,
                 editable_width: int = 90,
                 **kwargs):

        super().__init__(**kwargs)

        from vois.vuetify import settings

        self.on_change = on_change

        self.editable = editable
        self.editable_width = editable_width

        self.min_value = min_value
        self.max_value = max_value
        self.show_percentage = show_percentage
        self.decimals = decimals

        self._color = color
        if self._color is None:
            self._color = settings.color_first

        if max_int is None:
            if self.decimals <= 1:
                self.max_int = 10
            elif self.decimals == 2:
                self.max_int = 100
            elif self.decimals == 3:
                self.max_int = 1000
            elif self.decimals == 4:
                self.max_int = 10000
            else:
                self.max_int = 100000
        else:
            self.max_int = max_int

        self.postchar = ''
        if self.show_percentage:
            self.postchar = '%'

        self.label_width = label_width
        self.slider_width = slider_width
        self.reset_button = reset_button

        self.int_initial_value = self.float2integer(value)

        # self.label  = label.label(text, textweight=400, margins=0, margintop=4, height=22)

        if self.label_width > 0:
            spx = '%dpx' % self.label_width
            style = 'width: %s; min-width: %s; max-width: %s;' % (spx, spx, spx)
        else:
            style = ''
        self.label = v.Html(tag='div', children=[text], class_='pa-0 ma-0 mt-4', style_=style)
        self.slider = v.Slider(v_model=self.int_initial_value,
                               dense=True, xsmall=True,
                               ticks=False, thumb_size=10, dark=settings.dark_mode,
                               color=self._color, track_color="grey",
                               class_="pa-0 ma-0 ml-5 mr-4 mt-3 mb-n1",
                               style_='max-width: %dpx; width: %dpx;' % (self.slider_width, self.slider_width),
                               min=0, max=self.max_int, vertical=False, height=32)
        self.slider.on_event('input', self.on_input)
        self.slider.on_event('change', self.on_slider_change)

        self.field_value = v.TextField(autofocus=True, hide_details=True, single_line=True, hide_spin_buttons=False,
                                       dense=True, outlined=True, color=settings.color_first, type="number",
                                       class_='pa-0 ma-0 mt-2')
        self.c_field_value = v.Card(flat=True, children=[self.field_value], width=self.editable_width,
                                    max_width=self.editable_width)
        self.field_value.on_event('change', self.on_change_field_value)
        self.field_value.on_event('blur', self.on_change_field_value)

        if self.show_percentage:
            self.label_value = v.Html(tag='div', children=[str(self.int_initial_value) + self.postchar],
                                      class_='pa-0 ma-0 mt-4')
        else:
            self.label_value = v.Html(tag='div',
                                      children=['{:.{prec}f}'.format(value, prec=self.decimals) + self.postchar],
                                      class_='pa-0 ma-0 mt-4')

        self.label_value.on_event('click', self.on_value_change)

        if self.reset_button:
            self.bup = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0',
                             children=[v.Icon(color='grey', children=['mdi-menu-right'])])
            self.bup.on_event('click', self.on_up)
            self.cup = v.Card(children=[self.bup], elevation=0, class_='pa-0 ma-0 mr-1 mt-4',
                              style_="overflow: hidden;")
            if show_tooltip:
                self.cup = tooltip.tooltip("Increase", self.cup)

            self.bdn = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0',
                             children=[v.Icon(color='grey', children=['mdi-menu-left'])])
            self.bdn.on_event('click', self.on_dn)
            self.cdn = v.Card(children=[self.bdn], elevation=0, class_='pa-0 ma-0 mr-1 mt-4',
                              style_="overflow: hidden;")
            if show_tooltip:
                self.cdn = tooltip.tooltip("Decrease", self.cdn)

            self.bres = v.Btn(icon=True, x_small=True, rounded=False, elevation=0, width=15, height=15,
                              class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-close'])])
            self.bres.on_event('click', self.on_reset)
            self.cres = v.Card(children=[self.bres], elevation=0, class_='pa-0 ma-0 mr-1 mt-4',
                               style_="overflow: hidden;")
            if show_tooltip:
                self.cres = tooltip.tooltip("Reset value", self.cres)

            self.buttons = widgets.HBox([self.cdn, self.cres, self.cup])
        else:
            self.bup = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0',
                             children=[v.Icon(color='grey', children=['mdi-menu-up'])])
            self.bup.on_event('click', self.on_up)
            self.cup = v.Card(children=[self.bup], elevation=0, class_='pa-0 ma-0 mr-1 mt-2',
                              style_="overflow: hidden;")
            if show_tooltip:
                self.cup = tooltip.tooltip("Increase", self.cup)

            self.bdn = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0',
                             children=[v.Icon(color='grey', children=['mdi-menu-down'])])
            self.bdn.on_event('click', self.on_dn)
            self.cdn = v.Card(children=[self.bdn], elevation=0, class_='pa-0 ma-0 mr-1 mt-n1',
                              style_="overflow: hidden;")
            if show_tooltip:
                self.cdn = tooltip.tooltip("Decrease", self.cdn)

            self.buttons = widgets.VBox([self.cup, self.cdn])

        self.justify = 'start'
        self.class_ = 'pa-0 ma-0'
        self.no_gutters = True
        self.children = [self.label, self.slider, self.buttons, self.label_value]
        self.style_ = "overflow: hidden;"

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

        # When carriage return on the self.field_value or an external click

    def on_change_field_value(self, *args):
        self.row.children = [self.label, self.slider, self.buttons, self.label_value]
        self.value = float(self.field_value.v_model)
        self.field_value.v_model = self.value

    # On click on the self.labelvalue: display the self.field_value
    def on_value_change(self, *args):
        if self.editable and not self.slider.disabled:
            self.field_value.v_model = self.value
            self.row.children = [self.label, self.slider, self.buttons, self.c_field_value]

    # Input event on the slider
    def on_input(self, *args):
        if self.show_percentage:
            value = self.slider.v_model
            self.label_value.children = [str(value) + self.postchar]
        else:
            value = self.integer2float(self.slider.v_model)
            self.label_value.children = ['{:.{prec}f}'.format(value, prec=self.decimals) + self.postchar]

    # Input change on the slider
    def on_slider_change(self, *args):
        if self.show_percentage:
            value = self.slider.v_model
            self.label_value.children = [str(value) + self.postchar]
        else:
            value = self.integer2float(self.slider.v_model)
            self.label_value.children = ['{:.{prec}f}'.format(value, prec=self.decimals) + self.postchar]

        if not self.on_change is None:
            self.on_change(self.integer2float(self.slider.v_model))
        self.bup.disabled = self.slider.v_model >= self.max_int
        self.bdn.disabled = self.slider.v_model <= 0

    # Click on the up button
    def on_up(self, *args):
        if self.slider.v_model < self.max_int:
            self.slider.v_model = self.slider.v_model + 1
            self.on_slider_change()

    # Click on the down button
    def on_dn(self, *args):
        if self.slider.v_model > 0:
            self.slider.v_model = self.slider.v_model - 1
            self.on_slider_change()

    def on_reset(self, *args):
        self.slider.v_model = self.int_initial_value
        self.on_slider_change()

    # Conversion of float value [minvalue,maxvalue] to integer in [0,self.maxint]
    def float2integer(self, value):
        return int(self.max_int * (max(self.min_value, min(self.max_value, float(value))) - self.min_value) / (
                self.max_value - self.min_value))

    # Conversion of integer value [0,self.maxint] to float in [0.0,1.0]
    def integer2float(self, value):
        return ((self.max_value - self.min_value) * float(value)) / float(self.max_int) + self.min_value

    # Draw the widget
    def draw(self):
        """Returns the ipyvuetify object to display (a v.Row widget)"""
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self

    # Get the slider value
    @property
    def value(self):
        """
        Get/Set the slider value.

        Returns
        --------
        value : float
            Current value of the slider

        Example
        -------
        Programmatically set the slider value

            s.value = 0.56

        """
        return self.integer2float(self.slider.v_model)

    # Set the slider value
    @value.setter
    def value(self, new_value):
        if new_value < self.min_value:
            new_value = self.min_value
        if new_value > self.max_value:
            new_value = self.max_value

        int_value = self.float2integer(new_value)
        self.slider.v_model = int_value
        self.on_slider_change()

    @property
    def color(self):
        """
        Get/Set the widget color.

        Returns
        --------
        c : str
            widget color

        Example
        -------
        Programmatically change the widget color::

            s.color = '#00FF00'
            print(s.color)

        """
        return self._color

    @color.setter
    def color(self, c):
        if isinstance(c, str):
            self._color = c
            self.slider.color = self._color



"""Widget to display text strings vertically aligned."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v

from vois.vuetify.utils.util import *
from typing import Optional


#####################################################################################################################################################
# Testlist control
#####################################################################################################################################################
class TextList(v.Card):
    """
    Widget to vertically display a list of titles and texts strings. Each couple of title and text occupies a row.
        
    Parameters
    ----------
    titles : list of strings
        Strings to be displayed as title of each row
    texts : list of strings
        Strings to be displayed as the content of each row
    titles_bold : list of strings, optional
        List of titles whose corresponding texts should be displayed in with bold font (default is [])
    title_font_size : int, optional
        Size in pixel of the font used for the titles (default is 12)
    text_font_size : int, optional
        Size in pixel of the font used for the texts (default is 12)
    title_column : int, optional
        Number of column (out of 12) occupied by the titles (default is 4)
    text_column : int, optional
        Number of column (out of 12) occupied by the texts (default is 8)
    title_color : str, optional
        Color to use for the titles (default is 'black')
    text_color : str, optional
        Color to use for the texts (default is 'black')
    line_height_factor : float, optional
        Factor to multiply to the font-size to calculate the height of each row (default is 1.5)

    Example
    -------
    Creation and display of a widget to display some textual information::
        
        from vois.vuetify import TextList
        from IPython.display import display
        
        t = TextList(['Name', 'Surname', 'Address', 'Role'],
                              ['Davide', 'De Marchi', 'via Eduardo 34, Roccacannuccia (PE)', 'Software developer'],
                              titles_bold=['Surname'],
                              title_font_size=14,
                              text_font_size=16,
                              title_column=3,
                              text_column=10,
                              title_color='#003300',
                              text_color='#000000',
                              line_height_factor=1.4
                             )

        # Set some attributes of the card widget (margins, colors, width, etc.)
        t.class_    = 'pa-0 ma-4 ml-6 mr-8'
        t.flat      = False
        t.color     = '#e0ffe0'
        t.elevation = 8
        t.width     = '430px'

        display(t)


    .. figure:: figures/textlist.png
       :scale: 100 %
       :alt: textlist widget

       Textlist widget for displaying textual information.
   """

    deprecation_alias = dict(titlesbold='titles_bold', titlefontsize='title_font_size', textfontsize='text_font_size',
                             titlecolumn='title_column', textcolumn='text_column', titlecolor='title_color',
                             textcolor='text_color', lineheightfactor='line_height_factor')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 titles: list[str],
                 texts: list[str],
                 titles_bold: Optional[list[str]] = [],
                 title_font_size: int = 12,
                 text_font_size: int = 12,
                 title_column: int = 4,
                 text_column: int = 8,
                 title_color: str = 'black',
                 text_color: str = 'black',
                 line_height_factor: float = 1.5,
                 **kwargs):
        self.title_font_size = title_font_size
        self.text_font_size = text_font_size
        self.title_column = title_column
        self.text_column = text_column
        self.title_color = title_color
        self.text_color = text_color
        self.line_height_factor = line_height_factor
        self.div = v.Divider(class_='pa-0 ma-0 ml-1 mr-1', vertical=False)
        self.update(titles, texts, titles_bold)
        self.card = self

        super().__init__(flat=True, **kwargs)

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Update the widget with new titles and strings
    def update(self, titles, texts, titles_bold=[]):
        self.children = []
        children = []
        lineheight = "line-height: %dpx;" % (int(self.line_height_factor * (
            max(self.title_font_size, self.text_font_size))))  # To ensure vertical center alignment
        for title, text in zip(titles, texts):
            c1 = v.Col(cols=self.title_column, class_="pa-0 ma-0 ml-3",
                       children=[v.Html(tag='div', class_='pa-0 ma-0', children=[str(title)],
                                        style_='color: %s; font-weight: 450; text-transform: none; font-size: %dpx; %s' % (
                                            self.title_color, self.title_font_size, lineheight))],
                       style_="overflow: hidden;")

            if title in titles_bold:
                weight = 500
            else:
                weight = 350

            c2 = v.Col(cols=self.text_column, class_="pa-0 ma-0",
                       children=[v.Html(tag='div', class_='pa-0 ma-0',
                                        children=[str(text)],
                                        style_='color: %s; font-weight: %d; text-transform: none; font-size: %dpx; %s' % (
                                            self.text_color, weight, self.text_font_size, lineheight))],
                       style_="overflow: hidden;")
            children.append(v.Layout(class_="pa-0 ma-0", children=[c1, c2], style_="overflow: visible;"))
            children.append(self.div)

        self.children = children

    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Card widget)"""
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self



"""Widget to select among alternative options using a list of buttons displayed horizontally or vertically."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from vois.vuetify import Button
from vois.vuetify.utils.util import *
from typing import Callable, Optional


#####################################################################################################################################################
# Toggle buttons control
#####################################################################################################################################################
class Toggle(v.Row):
    """
    Widget to select among alternative options using a list of buttons displayed horizontally or vertically.
        
    Parameters
    ----------
    index : int
        Index of the selected option at start (from 0 to len(labels)-1)
    labels : list of strings
        Strings to be displayed as text of the options
    tooltips : list of strings, optional
        Tooltip text for the options (default is the empty list)
    icons : list of strings, optional
        Icons for the options (default is the empty list)
    on_change : function, optional
        Python function to call when the user clicks on one of the buttons. The function will receive a parameter of type int containing the index of the clicked button, from 0 to len(labels)-1
    row : bool, optional
        Flag to display the buttons horizontally or vertically (default is True)
    width : int, optional
        Width in pixels of the buttons (default is 150)
    height : int, optional
        Height in pixels of the buttons (default is 36)
    justify : str, optional
        In case of horizontal placement, applies the justify-content css property. Available options are: start, center, end, space-between and space-around.
    rounded : bool, optional
        Flag to display the buttons with a rounded shape (default is the button_rounded flag defined in the settings.py module)
    outlined : bool, optional
        Flag to display the buttons as outlined (default is False)
    color_selected : str, optional
        Color used for the buttons when they are selected (default is settings.color_first)
    color_unselected : str, optional
        Color used for the buttons when they are not selected (default is settings.color_second)
    dark : bool, optional
        Flag that controls the color of the text and icons in foreground (if True, the text and icons will be displayed in white, otherwise in black)
    padding_row : int, optional
        Horizontal padding among toggle buttons (1 unit means 4 pixels). Default is 1
    padding_col : int, optional
        Vertical padding among toggle buttons (1 unit means 4 pixels). Default is 2
    tile : bool, optional
        Flag to remove the buttons small border (default is False)
    large : bool, optional
        Flag that sets the large version of the button (default is False)
    x_large : bool, optional
        Flag that sets the xlarge version of the button (default is False)
    small : bool, optional
        Flag that sets the small version of the button (default is False)
    x_small : bool, optional
        Flag that sets the xsmall version of the button (default is False)

    Example
    -------
    Creation and display of a widget for the selection among 3 options::
        
        from vois.vuetify import Toggle
        from ipywidgets import widgets
        from IPython.display import display
        
        output = widgets.Output()

        def onchange(index):
            with output:
                print(index)

        t = Toggle(0, ['Option 1', 'Option 2', 'Option 3'], tooltips=['Tooltip for option 1'],
                          on_change=onchange, row=False, width=150, rounded=False)

        display(t)
        display(output)


    .. figure:: figures/toggle.png
       :scale: 100 %
       :alt: toggle widget

       Toogle widget for selecting alternative options using buttons.
   """

    deprecation_alias = dict(onchange='on_change', colorselected='color_selected', colorunselected='color_unselected',
                             paddingrow='padding_row', paddingcol='padding_col', xsmall='x_small',
                             xlarge='x_large')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 index: int,
                 labels: list[str],
                 tooltips: list[str] = [],
                 icons: list[str] = [],
                 on_change: Optional[Callable[[int], None]] = None,
                 row: bool = True,
                 width: int = 150,
                 height: int = 36,
                 justify: str = 'space-between',
                 rounded: Optional[bool] = None,
                 outlined: bool = False,
                 color_selected: Optional[str] = None,
                 color_unselected: Optional[str] = None,
                 dark: Optional[bool] = None,
                 padding_row: int = 1,
                 padding_col: int = 2,
                 tile: bool = False,
                 small: bool = False,
                 x_small: bool = False,
                 large: bool = False,
                 x_large: bool = False,
                 **kwargs):

        super().__init__(class_="pa-0 ma-0", justify=self.justify, style_="overflow: hidden;", **kwargs)

        from vois.vuetify import settings

        self.index = index  # Index of the selected button
        self.labels = labels
        self.tooltips = tooltips
        self.icons = icons
        self.on_change = on_change
        self.row = row
        self.width = width
        self.height = height
        self.justify = justify
        self.rounded = rounded if rounded is not None else settings.button_rounded
        self.outlined = outlined
        self._color_selected = color_selected if color_selected is not None else settings.color_first
        self._color_unselected = color_unselected if color_unselected is not None else settings.color_second
        self._dark = dark if dark is not None else settings.dark_mode
        self.padding_row = padding_row
        self.padding_col = padding_col
        self.tile = tile
        self.small = small
        self.x_small = x_small
        self.large = large
        self.x_large = x_large

        self.__createButtons()

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Create the toggle buttons
    def __createButtons(self):

        icon_color = 'black'
        if self._dark:
            icon_color = 'white'

        self.buttons = []
        i = 0
        for label in self.labels:
            tooltip = ''
            if i < len(self.tooltips):
                tooltip = self.tooltips[i]

            icon = None
            if i < len(self.icons):
                icon = self.icons[i]

            if self.row:
                c = "pa-0 ma-0 mr-%d" % self.padding_row
            else:
                c = "pa-0 ma-0 mb-%d" % self.padding_col

            b = Button(label, dark=self._dark, class_=c, small=self.small, xsmall=self.x_small, large=self.large,
                       xlarge=self.x_large,
                       icon=icon, icon_color=icon_color,
                       on_click=self.__internal_onchange, argument=i, width=self.width, height=self.height,
                       tooltip=tooltip, selected=(i == self.index),
                       rounded=self.rounded, tile=self.tile, outlined=self.outlined,
                       color_selected=self._color_selected,
                       color_unselected=self._color_unselected)
            self.buttons.append(b)
            i += 1

        if self.row:
            self.children = self.buttons
        else:
            self.col = v.Col(cols=12, class_="pa-0 ma-0", children=[x for x in self.buttons],
                             style_="overflow: hidden;")

            self.children = [self.col]

    # Get the active button
    @property
    def value(self):
        """
        Get/Set the active option index.
        
        Returns
        --------
        index : int
            Index of the selected option (from 0 to len(labels)-1)

        Example
        -------
        Programmatically select one of the options and print the index of the selected option::
            
            t.value = 2
            print(t.value)
        
        """
        return self.index

    # Set the active button
    @value.setter
    def value(self, index):
        if index >= 0 and index < len(self.buttons):
            self.buttons[self.index].selected = False
            self.index = index
            self.buttons[self.index].selected = True
            if self.on_change:
                self.on_change(self.index)

    # Get/Set the color of the selected button
    @property
    def color_selected(self):
        """
        Get/Set the color of the selected button.
        
        Returns
        --------
        color : str
            Color of the selected button

        Example
        -------
        Programmatically change the color of the selected button::
            
            t.colorselected = '#ff0000'
            print(t.colorselected)
        
        """
        return self._color_selected

    @color_selected.setter
    def color_selected(self, color):
        for b in self.buttons:
            b.color_selected = color
            if b._selected:
                b.b.color = color

    # Get/Set the color of the unselected button
    @property
    def color_unselected(self):
        """
        Get/Set the color of the unselected buttons.
        
        Returns
        --------
        color : str
            Color of the unselected buttons

        Example
        -------
        Programmatically change the color of the unselected buttons::
            
            t.colorunselected = '#ff0000'
            print(t.colorunselected)
        
        """
        return self._color_unselected

    @color_unselected.setter
    def color_unselected(self, color):
        for b in self.buttons:
            b.color_unselected = color
            if not b._selected:
                b.b.color = color

    # Get/Set the dark mode
    @property
    def dark(self):
        """
        Get/Set the dark mode of the buttons.
        
        Returns
        --------
        flag : bool
            If True the text and the icons are displayed in white color, otherwise in black color

        Example
        -------
        Programmatically change the dark mode::
            
            t.dark = Fals
            print(t.dark)
        
        """
        return self._dark

    @dark.setter
    def dark(self, flag):
        self._dark = flag

        for b in self.buttons:
            b.dark = self._dark

    # Manage onchange event
    def __internal_onchange(self, index):
        self.buttons[self.index].selected = False
        self.index = index
        self.buttons[self.index].selected = True
        if self.on_change:
            self.on_change(self.index)

    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Row or v.Col widget)"""
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self



"""Widget to select among alternative display using a list of tabs displayed horizontally or vertically."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v

from vois.vuetify import tooltip

from vois.vuetify.utils.util import *

from typing import Callable, Optional, Any


#####################################################################################################################################################
# Tabs control
#####################################################################################################################################################
class Tabs(v.Tabs):
    """
    Widget to select among alternative display using a list of tabs displayed horizontally or vertically.
        
    Parameters
    ----------
    index : int
        Index of the selected option at start (from 0 to len(labels)-1)
    labels : list of strings
        Strings to be displayed as text of the options
    contents : list of widgets, optional
        Widgets to be alternatively displayed when each of the tabs option is selected (for instance could be a list of ipywidgets.Output widgets). Default is None
    tooltips : list of strings, optional
        List of strings to be used as tooltips for the single tabs, in order (default is None)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
    on_change : function, optional
        Python function to call when the user clicks on one of the tabs. The function will receive a parameter of type int containing the index of the selected tab, from 0 to len(labels)-1
    row : bool, optional
        Flag to display the tabs horizontally or vertically (default is True)
            
    Example
    -------
    Creation and display of a tabs widget to select among alternative Outputs display::
        
        from vois.vuetify import Tabs
        from ipywidgets import widgets
        from IPython.display import display

        debug = widgets.Output()

        output0 = widgets.Output()
        output1 = widgets.Output()
        output2 = widgets.Output()

        with output0: print('This is output 0')
        with output1: print('This is output 1')
        with output2: print('This is output 2')

        def on_change(index):
            with debug:
                print(index)

        t = Tabs(0, ['Option 0', 'Option 1', 'Option 2'],
                      contents=[output0,output1,output2],
                      on_change=on:change, row=False)

        display(t)
        display(debug)

    .. figure:: figures/tabs.png
       :scale: 100 %
       :alt: tabs widget

       Creation of a tabs widget to display ipywidgets.Output content at every selection.
   """

    # Initialization
    deprecation_alias = dict(onchange='on_change')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 index: int,
                 labels: list[str],
                 contents: Optional[list[Any]] = None,
                 tooltips: Optional[list[str]] = None,
                 color: Optional[str] = None,
                 dark: Optional[bool] = None,
                 on_change: Optional[Callable[[int], None]] = None,
                 row: bool = True,
                 **kwargs):

        from vois.vuetify import settings

        self.dark = dark if dark is not None else settings.dark_mode

        self.index = index
        self.labels = labels
        self.on_change = on_change

        self._color = color
        if self._color is None:
            self._color = settings.color_first

        s = ''
        if self.dark:
            s = "color: %s;" % settings.textcolor_dark

        self.tab_list = []
        self.tab_list_with_tooltips = []
        i = 0
        for index, label in enumerate(self.labels):
            t = v.Tab(children=[label], style_=s, disabled=False)
            t.on_event('click', self.__internal_onchange)
            if isinstance(tooltips, list) and len(tooltips) > index:
                self.tab_list_with_tooltips.append(tooltip.tooltip(tooltips[index], t))
            else:
                self.tab_list_with_tooltips.append(t)

            self.tab_list.append(t)
            i += 1

        content_list = []
        if not contents is None:
            content_list = [v.TabItem(children=[x]) for x in contents]

        self.slider = v.TabsSlider(color=self._color, dark=self.dark)

        backcolor = 'white'
        textcolor = settings.textcolor_notdark
        if self.dark:
            backcolor = settings.dark_background
            textcolor = settings.textcolor_dark

        super().__init__(v_model=self.index, vertical=not row, dense=True, class_='pa-0 ma-0',
                         # background_color='white', color=settings.textcolor_notdark,
                         background_color=backcolor, color=textcolor,
                         children=[self.slider] + self.tab_list_with_tooltips + content_list,
                         **kwargs)

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Manage onchange event
    def __internal_onchange(self, widget, event, data):
        self.index = self.tab_list.index(widget)
        if self.on_change:
            self.on_change(self.index)

    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Tabs widget)"""
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self

    # Get the active tab
    @property
    def value(self):
        """
        Get/Set the active tab index.
        
        Returns
        --------
        index : int
            Index of the selected tab (from 0 to len(labels)-1)

        Example
        -------
        Programmatically select one of the tab and print the index of the selected option::
            
            t.value = 2
            print(t.value)
        
        """
        return self.index

    # Set the active button
    @value.setter
    def value(self, index):
        if index >= 0 and index < len(self.labels):
            self.index = index
            self.v_model = self.index
            if self.on_change:
                self.on_change(self.index)

    # disabled property
    @property
    def disabled(self):
        """
        Get/Set the disabled state.
        """
        return max([x.disabled for x in self.tab_list])

    @disabled.setter
    def disabled(self, flag):
        for tab in self.tab_list:
            tab.disabled = flag

    @property
    def color(self):
        """
        Get/Set the widget color.
        
        Returns
        --------
        c : str
            widget color

        Example
        -------
        Programmatically change the widget color::
            
            s.color = '#00FF00'
            print(s.color)
        
        """
        return self._color

    @color.setter
    def color(self, c):
        if isinstance(c, str):
            self._color = c
            self.slider.color = self._color



"""Widget to select independent options using a list of buttons displayed horizontally or vertically."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v

from vois.vuetify import Button

from vois.vuetify.utils.util import *

from typing import Callable, Optional


#####################################################################################################################################################
# Multi switch  control
#####################################################################################################################################################
class MultiSwitch(v.Html):
    """
    Widget to select independent options using a list of buttons displayed horizontally or vertically.
        
    Parameters
    ----------
    values : list of bool
        Initial state of the buttons representing the independent options
    labels : list of strings
        Strings to be displayed as text of the options
    tooltips : list of strings, optional
        Tooltip text for the options
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
    on_change : function, optional
        Python function to call when the user clicks on one of the buttons. The function will receive a parameter of list containing the status of all the buttons, from 0 to len(labels)-1
    row : bool, optional
        Flag to display the buttons horizontally or vertically (default is True)
    width : int, optional
        Width in pixels of the buttons (default is 150)
    height : int, optional
        Height in pixels of the buttons (default is 36)
    justify : str, optional
        In case of horizontal placement, applies the justify-content css property. Available options are: start, center, end, space-between and space-around.
    rounded : bool, optional
        Flag to display the buttons with a rounded shape (default is the button_rounded flag defined in the settings.py module)
    outlined : bool, optional
        Flag to display the buttons as outlined (default is True)
    color_selected : str, optional
        Color used for the buttons when they are selected (default is settings.color_first)
    color_unselected : str, optional
        Color used for the buttons when they are not selected (default is settings.color_second)
    manage_dbl_click : bool, optional
        If True the dblclick event is managed to select a single button of the multi-switch (default is False)
    padding_row : int, optional
        Horizontal padding among toggle buttons (1 unit means 4 pixels). Default is 1
    padding_col : int, optional
        Vertical padding among toggle buttons (1 unit means 4 pixels). Default is 2
    tile : bool, optional
        Flag to remove the buttons small border (default is False)
    large : bool, optional
        Flag that sets the large version of the button (default is False)
    x_large : bool, optional
        Flag that sets the x_large version of the button (default is False)
    small : bool, optional
        Flag that sets the small version of the button (default is False)
    x_small : bool, optional
        Flag that sets the x_small version of the button (default is False)

    Example
    -------
    Creation and display of a widget for the selection of 3 independent options::
        
        from vois.vuetify import MultiSwitch
        from ipywidgets import widgets
        from IPython.display import display
        
        output = widgets.Output()

        def on_change(values):
            with output:
                print(values)

        m = MultiSwitch([False, True, False], ['Option 1', 'Option 2', 'Option 3'],
                                    tooltips=['Tooltip for option 1'], 
                                    on_change=on_change, row=False, width=150, rounded=False, outlined=True,
                                    color_selected='#FFA300', color_unselected='#aaaaaa')

        display(m)
        display(output)


    .. figure:: figures/multiSwitch.png
       :scale: 100 %
       :alt: multiSwitch widget

       multiSwitch widget for selecting independent options using buttons.
   """

    # Initialization
    deprecation_alias = dict(onchange='on_change', colorselected='color_selected', colorunselected='color_unselected',
                             managedblclick='manage_dbl_click', paddingrow='padding_row', paddingcol='padding_col',
                             xsmall='x_small', xlarge='x_large')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 values: list[bool],
                 labels: list[str],
                 tooltips: list[str],
                 color: Optional[str] = None,
                 on_change: Optional[Callable[[list[bool]], None]] = None,
                 dark: Optional[bool] = None,
                 row: bool = True,
                 width: int = 150,
                 height: int = 36,
                 justify: str = 'space-between',
                 rounded: Optional[bool] = None,
                 outlined: bool = False,
                 color_selected: Optional[str] = None,
                 color_unselected: Optional[str] = None,
                 manage_dbl_click: bool = False,
                 padding_row: int = 1,
                 padding_col: int = 2,
                 tile: bool = False,
                 small: bool = False,
                 x_small: bool = False,
                 large: bool = False,
                 x_large: bool = False,
                 **kwargs):

        from vois.vuetify import settings

        super().__init__(tag='div', **kwargs)

        # self.style_ = 'padding: 0px; margin: 0px; ' + self.style_

        self.values = [bool(x) for x in values]  # list of boolean values
        self.labels = labels
        self.tooltips = tooltips
        self.color = color if color is not None else settings.color_first
        self.dark = dark if dark is not None else settings.dark_mode
        self.on_change = on_change
        self.row = row
        self.width = width
        self.height = height
        self.justify = justify
        self.rounded = rounded if rounded is not None else settings.button_rounded
        self.outlined = outlined
        self.color_selected = color_selected if color_selected is not None else settings.color_first
        self.color_unselected = color_unselected if color_unselected is not None else settings.color_second
        self.manage_dbl_click = manage_dbl_click
        self.padding_row = padding_row
        self.padding_col = padding_col
        self.tile = tile
        self.small = small
        self.x_small = x_small
        self.large = large
        self.x_large = x_large

        self.__createButtons()

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Create the toggle buttons
    def __createButtons(self):

        self.buttons = []
        i = 0
        for label in self.labels:
            tooltip = ''
            if i < len(self.tooltips):
                tooltip = self.tooltips[i]

            if self.row:
                c = "pa-0 ma-0 mr-%d" % self.padding_row
            else:
                if i == len(self.labels) - 1:
                    c = "pa-0 ma-0 mb-7"
                else:
                    c = "pa-0 ma-0 mb-%d" % self.padding_col

            b = Button(label,
                       class_=c,
                       # text_color=self.color,
                       on_click=self.__internal_onchange,
                       on_dblclick=self.__internal_dblclick,
                       argument=i,
                       tooltip=tooltip,
                       selected=self.values[i],
                       small=self.small,
                       x_small=self.x_small,
                       large=self.large,
                       x_large=self.x_large,
                       width=self.width,
                       height=self.height,
                       rounded=self.rounded,
                       tile=self.tile,
                       outlined=self.outlined,
                       dark=self.dark,
                       color_selected=self.color_selected,
                       color_unselected=self.color_unselected)
            self.buttons.append(b)
            i += 1

        if self.row:
            self.group = v.Row(class_="pa-0 ma-0", justify=self.justify, children=[x for x in self.buttons],
                               style_="overflow: hidden;")
        else:
            self.group = v.Col(cols=12, class_="pa-0 ma-0", children=[x for x in self.buttons],
                               style_="overflow: hidden;")

        self.children = [self.group]

    # Get the active button
    @property
    def value(self):
        """
        Get/Set the status of the multiSwitch buttons.
        
        Returns
        --------
        flags : list of booleans
            Selected status of all the options

        Example
        -------
        Programmatically set the options and print the status of the multiSwitch buttons::
            
            t.value = [False, True, True]
            print(t.value)
        
        """
        return self.values

    # Set the status of the buttons
    @value.setter
    def value(self, values):
        if len(values) == len(self.values):
            self.values = [bool(x) for x in values]
            for i in range(len(self.values)):
                self.buttons[i].selected = self.values[i]
            if self.on_change:
                self.on_change(self.values)

    # Manage on_change event
    def __internal_onchange(self, index):
        self.values[index] = not self.values[index]
        self.buttons[index].selected = self.values[index]
        if self.on_change:
            self.on_change(self.values)

    # Manage dblclick event
    def __internal_dblclick(self, index):
        if self.manage_dbl_click:

            i = 0
            for b in self.buttons:
                if i == index:
                    self.values[i] = True
                    b.selected = True
                else:
                    self.values[i] = False
                    b.selected = False

                i += 1

            if self.on_change:
                self.on_change(self.values)

    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Row or v.Col widget)"""
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self
